import subprocess
import platform
from pathlib import Path
import time
import os
import stat
import re
import urllib.request


class UaDfuApp:
    def __init__(self, driver_guid, pid, vid=0x20B1):
        assert platform.system() in [
            "Darwin",
            "Linux",
            "Windows",
        ], f"Unsupported platform: {platform.system()}"

        self.driver_guid = driver_guid
        self.pid = pid
        self.vid = vid

        if platform.system() == "Windows":
            self.dfu_app = (
                Path(os.environ["PROGRAMFILES"])
                / "XMOS"
                / "tusbaudiosdk"
                / "bin"
                / "Debug"
                / "x64"
                / "dfucons.exe"
            )
            assert self.dfu_app.exists(), f"dfucons.exe not found in location: {self.dfu_app}"
        elif platform.system() in ["Darwin", "Linux"]:
            xmosdfu_path = Path(__file__).parent / "xmosdfu"
            if not xmosdfu_path.exists():
                os_dir = "macos" if platform.system() == "Darwin" else "linux"
                xmosdfu_url = f"http://intranet.xmos.local/projects/usb_audio_regression_files/xmosdfu/{os_dir}/xmosdfu"
                r = urllib.request.urlopen(xmosdfu_url)
                with open(xmosdfu_path, "wb") as f:
                    f.write(r.read())
                xmosdfu_path.chmod(stat.S_IREAD | stat.S_IWRITE | stat.S_IEXEC)
            self.dfu_app = xmosdfu_path

    def download(self, image_bin):
        if platform.system() in ["Darwin", "Linux"]:
            cmd = [self.dfu_app, f"{hex(self.pid)}", "--download", image_bin]
        elif platform.system() == "Windows":
            # Issue 122: need a delay before DFU download, otherwise dfucons can fail
            time.sleep(10)
            cmd = [self.dfu_app, "download", image_bin, f"-g{self.driver_guid}"]
        ret = subprocess.run(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            timeout=300,
        )
        assert ret.returncode == 0, f"Download of image {image_bin} failed (error {ret.returncode})\nstdout:\n{ret.stdout}"

    def upload(self, output_file):
        platform_str = platform.system()
        if platform_str in ["Darwin", "Linux"]:
            cmd = [self.dfu_app, f"{hex(self.pid)}", "--upload", output_file]
        elif platform_str == "Windows":
            cmd = [self.dfu_app, "upload", output_file, f"-g{self.driver_guid}"]
        ret = subprocess.run(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            timeout=300,
        )
        assert (
            ret.returncode == 0
        ), f"Upload image to {output_file} failed (error {ret.returncode})\n{ret.stdout}"

    def revert_factory(self):
        platform_str = platform.system()
        if platform_str in ["Darwin", "Linux"]:
            cmd = [self.dfu_app, f"{hex(self.pid)}", "--revertfactory"]
        elif platform_str == "Windows":
            cmd = [self.dfu_app, "revertfactory", f"-g{self.driver_guid}"]
        ret = subprocess.run(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            timeout=300,
        )
        assert ret.returncode == 0, f"Revert to factory failed (error {ret.returncode})\nstdout:\n{ret.stdout}"

    def get_bcd_version(self):
        if platform.system() == "Windows":
            # Initially the version number can be reported unchanged after
            # a DFU download, so wait briefly before checking
            time.sleep(5)

        timeout = 30
        for _ in range(timeout):
            time.sleep(1)

            if platform.system() == "Darwin":
                ret = subprocess.run(
                    ["system_profiler", "SPUSBDataType"],
                    capture_output=True,
                    check=True,
                    text=True,
                )
                current_pid = None
                current_vid = None
                for i, line in enumerate(ret.stdout.splitlines()):
                    if line.strip().startswith("Product ID:"):
                        current_pid = int(line.split()[2], 16)
                    if line.strip().startswith("Vendor ID:"):
                        current_vid = int(line.split()[2], 16)
                    if line.strip().startswith("Version"):
                        if current_pid == self.pid and current_vid == self.vid:
                            return line.split()[1].strip()
            elif platform.system() == "Linux":
                ret = subprocess.run(
                    ["lsusb", "-vd", f"{hex(self.vid)}:{hex(self.pid)}"],
                    capture_output=True,
                    check=True,
                    text=True,
                )
                for line in ret.stdout.splitlines():
                    if line.strip().startswith("bcdDevice"):
                        version_str = line.split()[1]
                        return version_str.strip()
            elif platform.system() == "Windows":
                ret = subprocess.run(
                    [self.dfu_app, "info", f"-g{self.driver_guid}"],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    text=True,
                    timeout=300,
                )
                bcd_re = rf"\d+\s+0x{self.vid:04X}\s+0x{self.pid:04X}\s+0x(?P<bcd>\d{{4}})\s+"
                match = re.search(bcd_re, ret.stdout)
                if match:
                    match_dict = match.groupdict()
                    bcd_ver = match_dict["bcd"]
                    return f"{bcd_ver[:2]}.{bcd_ver[2:]}"

        assert 0, f"Failed to get device version after {timeout}s"
